"""
@authors: George Bertsch and John Mehlhaff

This module defines several symbols used by the Hartree-Fock code.
It also manages reading and parsing the files for the Hartree-Fock code.

The goals are transparency and modularity:
  - The symbols defined are presented at the top of the file (initialized to
    None as a placeholder).
  - The routines to initialize these symbols are decomposed into a set of
    function calls that takes place behind the scenes.

To gain access to all the symbols defined in this module, simply include

  import hfsetup

  init(inFile)

As lines in your program. The argument to init(), inFile, is documented below.
"""

import os
import sys
import numpy as np
from clebsch import *
import pdb

##### Symbols defining the dimension of the HF space
Nshell = None # The number of shell states used.
Nbasis = None # The total number of basis states generated from the shells.
Ndens  = None # The total number of elements of a packed matrix in HF space.
Nblock = None # The number of nonzero blocks of a matrix in HF space.
Nblockp = None # The number of proton blocks in the matrix
##### Symbols to aid in indexing packed matrices
# Given a block number index in the range [0, Nblock)...
b2D    = None  # Returns the dimension of that nonzero block in HF space.
b2idx0 = None  # Returns the starting index of that block in a packed HF matrix.
b2idx01 = None # Returns the starting orbital of that block.

##### Symbols to define the HF hamiltonian
T  = None # The self-energy part, a packed matrix T[0:Ndens]
vv = None # The interaction part, a packed rank 4 "tensor" vv[0:Ndens, 0:Ndens]
vvm = None  # The interaction for m1 >0; m2 < 0
time_reverse = None  # Packed time_reversal[0:Ndens] operator
Qop = None  #  Q operator in packed form
Qflag = 'N' #  'N','L','U' for no constraint, fixed Lagrange constraint, constraint by projection to achieve <Q> 
lambdaq = 0.0  # Lagrange multiplier for Q field, only with 'L'
Qtarget = None # target value for <Q>, only with 'L'

##### Symbols concerning the HF solution
u    = None # The initial guess at the HF solution.
focc = None # Occupation probabilities for orbitals at finite beta
nocc = None # Occupation numbers ground state blocks
update_n = False # true to update nocc in the iteration
##### Gradient method control parameters
eta_u = None # Tunes the step-size for the solution (u) update.
eta_f = None # Tunes the step-size for the occupation number update.
eta  = None # Dummy copy of eta_u for backwards compatibility.
errmax  = None # Convergence criterion: variance of an off-diagonal element of Horb
itm  = None # The maximum number of allowed iterations.
##### Finite temperature parameters
betaL = None   # a list of beta's to calculate
Np = None # target number of protons
Nn = None # target number of neutrons
Nbasis_n = None # number of neutron orbitals
Nbn = None # number of neutron blocks
##### Symbols used by this file that may prove useful elsewhere but
##### are not used by external code at this time (07/15/15).
# Given a basis state index in the range [0, Nbasis)...
k2is = None # Returns the shell no. (line in the spsFile) the state is from.
k2el = None # Returns the orbital angular momentum of that state.
k2j  = None # Returns the total angular momentum of that state.
k2m  = None # Returns TWICE the azimuthal angular momentum of that state.
k2ch = None # Returns the charge of that state.
k2b  = None # returns the block of that state
o2b  = None # block of orbital
evalsL = None   # eigenvalues of H_orb
# Given a shell number in the range [0, Nshell)...
par_ch = None # Returns a tuple (par, ch) of the parity and charge of the shell.
sh2j   = None # Returns TWICE the total angular momentum of that shell.

Jmax = None # TWICE the largest total angular momentum of any shell.

# 'blocks' is a dictionary mapping (par, ch, twice azimuthal ang. mom.) to
# all the states in blocks possessing that parity, charge, and azimuthal angular
# momentum. 'keys' is a sorted list of the keys of 'blocks'.
blocks = None
keys   = None

# Given a shell index 'ish' in the range [0, Nshell) and an azimuthal angular
# momentum 'midx' in the range [0, (Jmax - 1) / 2], these arrays indexed at
# [ish, midx] will...
sm2k = None # Return the state index in [0, Nbasis) (use with the k2__ arrays)
sm2b = None # Return the block number of the state in [0, Nblock)
sm2i = None # Return the index of the state in blocks[keys[sm2b[ish, midx]]]

# Initializes all of the above symbols given an input file.
#
# @param: inFile - A string pointing to an input file for the Hartree-Fock
#     method.
# @requires: 'inFile' is formatted as in the SAMPLE input file. Note that it
#     is allowed for the lines specifying the initial guess file and the file
#     containing starting occupation numbers be specified as "None". This is
#     to accommodate code that may be written to construct these files but
#     that still needs to access information regarding the basis states and
#     Hamiltonian in their absence.
def init(inFile,GSorFT):
  spsFile, intFile, occFile, uFile, Qline, parLine = \
      parseInFile(inFile,GSorFT)

  # The user should be able to specify files relative to the location of
  # 'inFile' in the file system.
#  fileLoc = lambda fileName: os.path.join(os.path.dirname(inFile), fileName)
  readBasis(spsFile)
  initBlocks()
  initHamiltonian(intFile)
  if GSorFT == 'FT':
    foccFromFile(occFile)
  else:
    noccFromFile(occFile)
  parseQ(Qline)
  if uFile == 'None':
    Unitmatrix()
  else:
    uFromFile(uFile)
  initIterParams(parLine)

"""
The following routines comprise the "workhorses" of the init() function. Reading
them is not conceptually illuminating and calling them is not supported. Please
stick to the init() routine with the documentation presented above as your
guide.
"""
def readBasis(spsFile):
  global Nshell, Nbasis,evalsL
  global k2is, k2el, k2j, k2m, k2ch, par_ch, sh2j, Jmax
  global Nbasis_n
  lines = open(spsFile).readlines()
  Nshell = len(lines)
  k2is = []
  k2el = []
  k2j = []
  k2m= []
  k2ch = []
  Nbasis = 0
  Nbasis_n = 0
  Jmax = 0
  par_ch = [(-1,-1)]
  sh2j = [ -1]
  for ish in range(Nshell):
    ss = lines[ish].split()
    j2b=int(float(ss[3])*2)
    Nbasis += (j2b + 1)/2
    par = int(float(ss[2]))%2
    ch  = int(float(ss[4])+0.5)
    par_ch.append((par,ch))
    sh2j.append(j2b)
    for m in range(1,j2b+1,2):
      k2is.append(ish)
      k2m.append(m)
      k2j.append(j2b)
      if j2b>Jmax : Jmax = j2b
      k2el.append(int(float(ss[2])))
      k2ch.append(ch)  
      if ch == 0:
        Nbasis_n += 1
  k2is = np.array(k2is)
  k2m = np.array(k2m)
  k2j = np.array(k2j)
  k2el = np.array(k2el)
  k2ch = np.array(k2ch)
  evalsL = np.zeros(Nbasis)

def initBlocks():
  global Nblock, Nbn, Ndens
  global b2D, b2idx0, b2idx01
  global blocks, keys, sm2k, sm2b, sm2i, k2b,o2b

  blocks = {}
  for m2b in range(1,Jmax+1,2):
    for par in range(2):
      for ch in (1,0):
        block = []
        for k in range(Nbasis):
          if (k2m[k] == m2b) and (k2el[k]%2 == par) and (k2ch[k] == ch):
            block.append(k)
          if block != []: blocks[(ch,par,m2b)] = np.array(block)
  keys = blocks.keys()
  keys.sort()
  Nblock = len(blocks)
  Nbn = 0
  for key in keys:
    if key[0] == 0:  Nbn += 1
  b2idx0 = np.array([0]*(Nblock+1))
  b2D   = np.array([0]*Nblock)
  idx_start = 0
  Nbasis2 = 0
  for ib in range(Nblock):
    orbs_b = blocks[keys[ib]]
    Db = len(orbs_b)
    b2D[ib] = Db
    b2idx0[ib] = idx_start
    idx_start += len(orbs_b)**2
    Nbasis2 += Db
  Ndens = b2idx0[-1] = idx_start

  b2idx01 = np.zeros(Nblock + 1, dtype = int)
  b2idx01[1:] = np.cumsum(b2D)
  b2idx01[0] = 0
  k2b = np.array([0]*Nbasis)
  sm2k = np.array([[-1]*((Jmax+1)/2)]*Nshell)
  sm2b = np.array([[-1]*((Jmax+1)/2)]*Nshell)
  for k in range(Nbasis):
    if k2m[k] > 0 :
      ish= k2is[k]
      midx = (k2m[k]-1)/2
      par = k2el[k]%2
      ch = k2ch[k] 
      ib = keys.index((ch,par,k2m[k]))
      sm2k[ish,midx] =k
      sm2b[ish,midx] = ib
      k2b[k] = ib
  sm2i = np.array([[-1]*((Jmax+1)/2)]*Nshell)
  for ib in range(Nblock):
    states = blocks[keys[ib]]
    for i in range(len(states)):
      k = states[i]
      ish = k2is[k]
      midx = (k2m[k]-1)/2
      sm2i[ish,midx] = i
    idx= 0
    o2b = np.array([0]*Nbasis)
    for ib in range(Nblock):    
      ch,par,m = keys[ib]
      Db = b2D[ib]
      idx1 = idx+Db
      for ix in range(idx,idx1):
        o2b[ix] = ib
      idx = idx1

def mk_TR():
 global time_reverse
 for i1 in range(1,Nshell+1):
   for i3 in range(1,Nshell+1):
     j1=sh2j[i1]; j3=sh2j[i3]
     iz1 = i1-1; iz3 = i3-1
#     print i1,i3,par_ch[i1],par_ch[i3]
     if(par_ch[i1] == par_ch[i3]) :
        m1_max = min(j1,j3)
        for m1 in range(1,m1_max+1,2):
          midx1 = (m1-1)/2  
          ib1 = sm2b[iz1,midx1]
          ii1 = sm2i[iz1,midx1]
          ii3 = sm2i[iz3,midx1]
          ib3 = sm2b[iz3,midx1]
          if ib3 != ib1:
            print 'block problem', i,j,k,l
            exit()
          idx13 = b2idx0[ib1]+b2D[ib1]*ii1 + ii3
          time_reverse[idx13] = (-1)**((j1+j3)/2+1)

def mk_vv(lines, Nv):
 global vv, vvm
 quadruples = []
 print 'Nv',Nv
 for ii in range(Nv):     #Nv
  ss = lines[ii].split()
  # print ss
  i,j,k,l,J = map(int,ss[:5])
  left = [i,j]
  left.sort()
  right = [k,l]
  right.sort()
  quad = [left,right]
  quad.sort()
  quad.append(J)
  if (quad in quadruples):     # check for duplicate entries
    print 'repeated quad', quad
  else:
    quadruples.append(quad) 
  vj = float(ss[5])
  sqrt2 = 2.0**0.5
  if i == j:  vj = vj/sqrt2
  if k == l:  vj = vj/sqrt2
  J=J*2
  sgn2 = (-1)**((sh2j[i]+sh2j[j]+sh2j[k]+sh2j[l])/2)
  permutations = [(i,j,k,l,'as_is')]                     #as read it
  if left != right :
    permutations.append((k,l,i,j,'adjoiont'))   # adjoint
  for perm in permutations:
    i1,i2,i3,i4,adj  = perm
    vjj = vj
    for dx in ('direct','exchange'):
      iz1 = i1-1; iz2 = i2-1;iz3 = i3-1;iz4 = i4-1;
      if dx == 'exchange':
        it = i3; i3 = i4; i4 = it; iz4 = i4-1; iz3 = i3-1
        vjj = vjj*(-1)**((sh2j[i3]+sh2j[i4]+J)/2+1)
      j1=sh2j[i1]; j2=sh2j[i2]; j3=sh2j[i3]; j4=sh2j[i4]
      if(par_ch[i1] == par_ch[i3]) :        
        m1_max = min(j1,j3)
        for m1 in range(1,m1_max+1,2):
          midx1 = (m1-1)/2  
          ib1 = sm2b[iz1,midx1]
          ii1 = sm2i[iz1,midx1]
          ii3 = sm2i[iz3,midx1]
          ib3 = sm2b[iz3,midx1]
          if ib3 != ib1:
            print 'block problem', i,j,k,l
            exit()
          idx13 = b2idx0[ib1]+b2D[ib1]*ii1 + ii3
          m2_max = min(j2,j4)
          for m2 in range(1, m2_max+1,2):
            midx2 = (m2-1)/2 
            ib2 = sm2b[iz2,midx2]
            ii2 = sm2i[iz2,midx2]
            ii4 = sm2i[iz4,midx2]
            ib4 = sm2b[iz4,midx2]
            if ib4 != ib2:
              print 'block problem', i,j,k,l
              exit()
            idx24 = b2idx0[ib2]+b2D[ib2]*ii2 + ii4
            c1 = clebsch(j1,j2,J,m1,m2)   # plus m
            c2 = clebsch(j3,j4,J,m1,m2) 
            c3 = clebsch(j1,j2,J,m1,-m2)  # minus m
            c4 = clebsch(j3,j4,J,m1,-m2) 
            vpm = vjj*c1*c2                        # +m2
            vmm = vjj*c3*c4     #*(-1)**((j1+j2)/2+1)    # -m2
            vv[idx13,idx24] += vpm  
            vv[idx24,idx13] += vpm
            vvm[idx13,idx24] += vmm
            vvm[idx24,idx13] += vmm*sgn2
            dummy = 1
  dummy = 2



def initHamiltonian(intFile):
  global T, vv, vvm, time_reverse
#  fl = factor()

  vv = np.array([[0.0]*Ndens]*Ndens)
  time_reverse = np.array([0.0]*Ndens)
  vvm = vv*1.0
  lines = open(intFile).readlines()
 
  # Skip comments and white space.
  i = 0
  while (lines[i].startswith("#") or (lines[i] == "") or (lines[i].isspace())):
    i += 1

  ss = lines[i].split()
  Nv = int(ss[0])
  i += 1

  # Skip more potential comments and white space.
  while (lines[i].startswith("#") or (lines[i] == "") or (lines[i].isspace())):
    i += 1

  lines = lines[i:]

  # Initialize self-energies.
  eeL = np.array([0.0]*Nshell)
  for i in range(Nshell):
    eeL[i] = float(ss[i+1])
  T = np.array([0.0]*Ndens)
  for ib in range(Nblock):
    states = blocks[keys[ib]]
    Db = b2D[ib]
    for k in states:
      midx = (k2m[k]-1)/2
      ii = sm2i[k2is[k],midx]
      idx = b2idx0[ib]+b2D[ib]*ii+ii
      T[idx] = eeL[k2is[k]]
  # Initialize interactions.
  mk_vv(lines, Nv)
  mk_TR()

# Parses the occFile as returned by parseInFile() into an array of occupation
# probabilities focc[0:Nbasis]
def foccFromFile(occFile):
  global focc
  ifile = open(occFile, 'r')
  line = ifile.readlines()[0]
  ifile.close()
  focc = np.array(map(float, line.split()))
  if (len(focc) != Nbasis):
    print "%s did not specify occupation #'s for %d orbitals" % \
      (occFile, Nbasis)
    sys.exit(1)

def noccFromFile(occFile):
  global nocc
  ifile = open(occFile, 'r')
  line = ifile.readlines()[0]
  ifile.close()
  
  nocc = np.array(map(int, line.split()))
  if (len(nocc) != Nblock):
    print "%s did not specify occupation #'s for %d blocks" % \
      (occFile, Nblock)
    print "Aborting..."
    sys.exit(1)

# Parses the uFile as returned by parseInFile() into an array u[0:Ndens]
def uFromFile(uFile):
  global u
  u = np.loadtxt(uFile)

def initIterParams(paramLine):
  global eta_u, eta_f, eta, errmax, itm

  split = paramLine.split()

  if (len(split) < 3) :
    print "Expected 3 or 4 algorithm parameters.",
    print "Got %d: %s" % (len(split), paramLine)
    print "Aborting..."
    sys.exit(1)
  if (len(split) == 3):
    eta_u = float(split[0])
    errmax = float(split[1])
    itm = int(split[2])
  elif (len(split) == 4):
    eta_u = float(split[0])
    eta_f = float(split[1])
    errmax = float(split[2])
    itm = int(split[3])
  eta = eta_u


def parseQ(line):
  global Qop,Qflag,lambdaq,Qtarget
  ss = line.split()
  Qfile = ss[0]; Qflag = ss[1]
  if Qflag == 'L' : 
    lambdaq = float(ss[2])
  elif Qflag == 'C':
    Qtarget  = float(ss[2])
  f1 = open(Qfile, 'r')
  dlines = f1.readlines()
  f1.close()
  i = 0
  while (dlines[i].startswith("#")):
    i += 1
  Qop = np.loadtxt(Qfile, skiprows = i)

  if (len(Qop) != Ndens):
    print "%s did not specify an operator of the correct shape" % dfileName
    print "Expected %d entries, got %d" % (Ndens, len(operator))
    print "Aborting..."
    sys.exit(1)
  return Qop

# Returns a 6-tuple: (spsFile, intFile, occFile, uFile, paramLine,
#     constraintLines)
# Raises an IOError if 'inFile' does not have at least 5 non-comment lines
#     corresponding to the first 5 items above.
def parseInFile(inFile,GSorFT):
  global Np,Nn,betaL,update_n
  ifile = open(inFile, 'r')
  lines0 = ifile.readlines()
  ifile.close()
  lines = []
  for line in lines0:
    if not (line.startswith("#")): lines.append(line)
  i = 0
  spsFile = lines[0].split()[0]
  intFile = lines[1].split()[0]
  ss2 = lines[2].split()
  occFile = ss2[0]   
  if GSorFT == 'GS':
    if ss2[1] == 'U': update_n = True
  uFile   = lines[3].split()[0]
  Qline = lines[5]
  paramLine = lines[4]
  if GSorFT == 'FT':
    Np,Nn = map(float,lines[6].split()[:2])
    betaL = []
    for line in lines[7:]:
      betaL.append(float(line.split()[0]))
  return spsFile, intFile, occFile, uFile, Qline, paramLine
def Unitmatrix():
  global u
  u = np.array([0.0]*Ndens)
  for ib in range(Nblock):
    Db = b2D[ib]
    idty = np.array([[0.0]*Db]*Db)
    for id in range(Db):
      idty[id,id] = 1.0
    i1 = b2idx0[ib]
    i2 = b2idx0[ib + 1]
    u[i1:i2] = idty.ravel()
 