"""
Utilities that are general to both the zero and finite temperature cases.
"""

import numpy as np
import sys
import pdb

import hfsetup as hfs

# Numerical value for floating point comparisons.
FLOAT_EPS = 1e8 * sys.float_info.epsilon

#####
##### Routines for quickly unpacking/repacking data from/to packed matrices.
#####
# Packs the square array b into the region of array a corresponding to block
# ib. Note requirement 0 <= ib < Nblock. Also b must have the right dimensions
# to fit back into the ib-th block of a.
def packBlock(a, b, ib):
    if (ib < 0) or (ib >= hfs.Nblock):
        raise ValueError("Illegal block requested: %s" % ib)

    Db = hfs.b2D[ib]
    if (b.shape != (Db, Db)):
        raise ValueError("Array b cannot be packed into a.")

    i1 = hfs.b2idx0[ib]
    i2 = hfs.b2idx0[ib + 1]
    a[i1:i2] = b.ravel()

# Takes a packed array a[0:Ndens] and returns the square matrix corresponding
# to the ib-th block of a. Note requirement 0 <= ib < Nblock.
def unpackBlock(a, ib):
    if (ib < 0) or (ib >= hfs.Nblock):
        raise ValueError("Illegal block requested: %s" % ib)

    i1 = hfs.b2idx0[ib]
    i2 = hfs.b2idx0[ib + 1]
    Db = hfs.b2D[ib]

    # It turns out reshaping does not copy the data in memory.
    # Changing to a manual copy to avoid what was otherwise a subtle bug...
    return a[i1:i2].reshape(Db, Db).copy()

# Retrieve the pairwise occupation number differences (fa - fb)  within a block.
def occnumBlock(occnums, ib):
    if (ib < 0) or (ib >= hfs.Nblock):
        raise ValueError("Illegal block requested: %s" % ib)

    # Reshaping the occupation number array so that
    # occnumMatrix[i, j] = occnumb[i] - occnumb[j]
    Db = hfs.b2D[ib]
    occnumb = occnums[hfs.b2idx01[ib]:hfs.b2idx01[ib + 1]]
    occnumMatrix = np.tile(occnumb, (Db, 1))
    occnumMatrix = np.transpose(occnumMatrix) - occnumMatrix
    return occnumMatrix

# Transform an operator Osp from the basis to the orbital representation.
def calcOrbOp(Osp, u):
    Oorb = np.zeros(hfs.Ndens)
    for ib in range(hfs.Nblock):
        ub = unpackBlock(u, ib)
        Ospb = unpackBlock(Osp, ib)
        Oorbb = np.dot(ub, np.dot(Ospb, np.transpose(ub)))
        packBlock(Oorb, Oorbb, ib)
    return Oorb

##### Routines to calculate essential operators and quantities of interest.
def calcRho(u, occnums):
    rho = np.zeros(hfs.Ndens)
    ustar = np.conjugate(u)
    ibn = 0 # The index of the current block in 'occnums'
    for ib in range(hfs.Nblock):
        Db = hfs.b2D[ib]
        i0 = hfs.b2idx0[ib]
# Unpack the blocks we need.
        occnumb = np.diag(occnums[ibn:ibn + Db])
        uconjb  = np.transpose(unpackBlock(ustar, ib))
        ub      = unpackBlock(u, ib)
# Compute rho for this block and repack.
        rhob = np.dot(uconjb, np.dot(occnumb, ub))
        packBlock(rho, rhob, ib)
        ibn += Db
    return rho

def calcV(rho):
    V = np.dot(hfs.vv, rho) + np.dot(hfs.vvm, hfs.time_reverse * rho)
    return V

def calcT(rho):
    return hfs.T 

def calcHsp(rho):
  Hsp = hfs.T + calcV(rho)
  if hfs.lambdaq != 0.0:
    Hsp = Hsp + hfs.lambdaq*hfs.Qop
  return Hsp

def totalE(rho):
    E = np.dot(rho, calcT(rho))    *2  # to sum p/m m
    E += 0.5 * np.dot(rho, calcV(rho)) *2   # to sum p/m
    return E

def calcQ(rho):
    Qp = 2*np.dot(rho[:hfs.Nbasis_n],hfs.Qop[:hfs.Nbasis_n])    # factor 2 to sum p/m
    Qn = 2*np.dot(rho[hfs.Nbasis_n:],hfs.Qop[hfs.Nbasis_n:])
    return Qp+Qn 


# Verify that the packed matrix M is unitary.
def isUnitary(M):
    for ib in range(hfs.Nblock):
        Mb = unpackBlock(M, ib)
        idty = np.diag(np.ones(Mb.shape[0]))
        nearIdty = np.dot(np.transpose(Mb), Mb)
        allNegs = np.abs(nearIdty - idty) - FLOAT_EPS
        if not ( (allNegs <= 0.0).all() ):
            return False
    return True

#####
##### Printing routines to effectively visualize data.
#####
# Print the packed matrix M[0:Ndens]
def printPacked(M, label = None):
    if (label):
        print label + " = ..."

    for ib in range(hfs.Nblock):
        Db = hfs.b2D[ib]
        i0 = hfs.b2idx0[ib]
        block = M[i0:i0 + Db**2].reshape((Db, Db))
        print block

def packedTrace(a, b):     # matrices must be in packed form
    return np.dot(a, b)

def Op_ph(Aorb,focc):    #   multiply (fa-fb) Aorb_ab
  Aph = Aorb*0.0
  for ib in range(hfs.Nblock):
    Ab = unpackBlock(Aorb, ib)
    Aphb = Ab*0.0
    Aphb = Ab*occnumBlock(focc,ib)
    packBlock(Aph,Aphb,ib)
  return Aph

def get_diagonal(M):
    diags = []*hfs.Nbasis
    for ib in range(hfs.Nblock):
       M1 = unpackBlock(M,ib)
       Db = hfs.b2D[ib]
       for i in range(Db):
          diags.append(M1[i,i])
    return np.array(diags)   
